/*
 *  Database record class for DAAP sharing
 *
 *  Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gst/gst.h>
#include <wand/wand_api.h>
#include <libexif/exif-data.h>
#include <libexif/exif-content.h>
#include <libexif/exif-entry.h>

#include "dmapd-daap-record.h"
#include "g-gst-input-stream.h"
#include "g-gst-input-stream-factory.h"
#include "g-gst-wav-input-stream-factory.h"
#include "g-gst-mp3-input-stream-factory.h"
#include "g-gst-raw-input-stream-factory.h"
#include "util.h"

struct DmapdDAAPRecordPrivate {
	gint filesize;
	const char *location;
	const char *format;	 /* Format, possibly after transcoding. */
	const char *real_format;
	const char *title;
	const char *album;
	const char *artist;
	const char *genre;
	gboolean has_video;
	gint rating;
	gint32 duration;
	gint32 track;
	gint32 year;
	gint32 firstseen;
	gint32 mtime;
	gint32 disc;
	gint32 bitrate;
	gint transcode_format;
	GGstInputStreamFactory *factory;
};

GOnce once = G_ONCE_INIT;
GMutex *tag_read;
static GstElement *pipeline, *src, *decode, *typefind;

static void
newpad_cb (GstElement *decodebin, GstPad *pad, gboolean last, gpointer found)
{
	GstCaps *caps;
	const gchar *mimetype;
	GstStructure *structure;

	/* Get mimetype. */
	caps = gst_pad_get_caps (pad);
	if (caps == NULL) {
		g_warning ("Error getting caps from pad");
		return;
	}

	structure = gst_caps_get_structure (caps, 0);
	if (structure == NULL) {
		g_warning ("Error getting structure from caps");
		return;
	}

	mimetype = gst_structure_get_name (structure);

	if (g_strrstr (mimetype, "video")) {
		*(gboolean *) found |= TRUE;
	} else if (g_strrstr (mimetype, "audio")) {
		GstPad *typefind_pad;

		typefind_pad = gst_element_get_static_pad (typefind, "sink");
		g_assert (typefind_pad != NULL);

		if (pads_compatible (pad, typefind_pad)) {
			g_assert (! GST_PAD_IS_LINKED (
				gst_element_get_static_pad (typefind, "sink")));
			gst_pad_link (pad, typefind_pad);
		}
	}

	gst_caps_unref (caps);
}

static gboolean
message_loop (GstElement *element, GstTagList **tags)
{
	GstBus *bus;
	gboolean done = FALSE;

	bus = gst_element_get_bus (element);
	g_return_val_if_fail (bus != NULL, FALSE);
	g_return_val_if_fail (tags != NULL, FALSE);

	while (! done) {
		GstMessage *message;

		message = gst_bus_pop (bus);

		if (message == NULL)
			/* All messages read, we're done */
			break;

		switch (GST_MESSAGE_TYPE (message)) {
		case GST_MESSAGE_TAG: {
			GstTagList *new_tags, *old_tags;

			gst_message_parse_tag (message, &new_tags);
			if (*tags) {
				old_tags = *tags;
				*tags = gst_tag_list_merge (old_tags,
							    new_tags,
							    GST_TAG_MERGE_KEEP);
				gst_tag_list_free (old_tags);
			} else
				*tags = new_tags;
			break;
		}
		default:
			break;
		}
		gst_message_unref (message);
	}

	gst_object_unref (bus);

	return TRUE;
}

static gchar *
determine_format (DmapdDAAPRecord *record, const gchar *description)
{
	gchar *format;

	if (g_strrstr (description, "MP3"))
		format = g_strdup ("mp3");
	else if (g_strrstr (description, "MPEG-4 AAC"))
		format = g_strdup ("aac");
	else if (g_strrstr (description, "Vorbis"))
		format = g_strdup ("ogg");
	else if (g_strrstr (description, "FLAC"))
		format = g_strdup ("flac");
	else {
		gchar *ext;

		ext = strrchr (record->priv->location, '.');
		if (ext == NULL) {
			ext = "mp3";
		} else {
			ext++;
		}

		format = g_strdup (ext);
	}

	return format;
}

static void
insert_tag (const GstTagList * list, const gchar * tag, DmapdDAAPRecord *record)
{
	gint i, count;

	if (tag == NULL)
		return;

	count = gst_tag_list_get_tag_size (list, tag);

	for (i = 0; i < count; i++) {
		gchar *val;

		if (gst_tag_get_type (tag) == G_TYPE_STRING) {
			if (!gst_tag_list_get_string_index (list, tag, i, &val))
				g_assert_not_reached ();
		} else {
			val = g_strdup_value_contents (gst_tag_list_get_value_index (list, tag, i));
		}

		if (! strcmp ("title", tag)) {
			record->priv->title = val;
		} else if (! strcmp ("artist", tag)) {
			record->priv->artist = val;
		} else if (! strcmp ("album", tag)) {
			record->priv->album = val;
		} else if (! strcmp ("genre", tag)) {
			record->priv->genre = val;
		} else if (! strcmp ("audio-codec", tag)) {
			record->priv->real_format = determine_format (record, val);
			/* Determine the format to "advertise" (i.e., used for 
			 * URL path)
			 */
			if (record->priv->has_video) {
				/* FIXME: get from video stream. */
				gchar *ext;

				ext = strrchr (record->priv->location, '.');
				if (ext == NULL) {
					ext = "mov";
				} else {
					ext++;
				}
				record->priv->format = g_strdup (ext);
			} else {
				/* No transcoding unless transcode_format
				 * set later.
				 */
				record->priv->format = g_strdup (record->priv->real_format);
			}
		} else if (! strcmp ("track-number", tag)) {
			record->priv->track = atoi (val);
			g_free (val);
		} else {
			/* Unused. */
			g_free (val);
		}
	}
}

static void
setup_mutex (gpointer user_data)
{
	tag_read = g_mutex_new ();
}

static void
setup_pipeline (void)
{
	/* Set up pipeline. */
	pipeline = gst_pipeline_new ("pipeline");

	src = gst_element_factory_make ("filesrc", "source");
	if (src == NULL) {
		g_warning ("Error creating filesrc element");
		return;
	}

	decode = gst_element_factory_make ("decodebin", "decoder");
	if (decode == NULL) {
		g_warning ("Error creating decodebin element");
		return;
	}

	typefind = gst_element_factory_make ("typefind", "typefind");
	if (typefind == NULL) {
		g_warning ("Error creating typefind element");
		return;
	}

	gst_bin_add_many (GST_BIN (pipeline), src, decode, typefind, NULL);
	if (gst_element_link (src, decode) == FALSE) {
		g_warning ("Error linking pipeline");
		return;
	}
}

static void
teardown_pipeline (void)
{
	g_object_unref (pipeline);
}

static gboolean
pause_pipeline (GstElement *pipeline)
{
	GstState state;
	GstStateChangeReturn sret;

	sret = gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_PAUSED);
	if (GST_STATE_CHANGE_ASYNC == sret) {
		if (GST_STATE_CHANGE_SUCCESS != gst_element_get_state (GST_ELEMENT (pipeline), &state, NULL, 1 * GST_SECOND)) {
			g_warning ("State change failed");
		}
	} else if (sret != GST_STATE_CHANGE_SUCCESS) {
		g_print ("Could not read file");
	}

	/* Run once. */
	return FALSE;
}

static gboolean
quit_mainloop (GMainLoop *loop)
{
	g_main_loop_quit (loop);

	/* Run once. */
	return FALSE;
}

static void
no_more_pads_cb (GstElement *element, GMainLoop *loop)
{
	quit_mainloop (loop);	
}

static void
read_metadata (DmapdDAAPRecord *record, const gchar *path)
{
	GstFormat fmt;
	GstTagList *tags = NULL;
	gint64 nanoduration;
	gboolean has_video = FALSE;
	GMainLoop *loop;

	g_once (&once, (GThreadFunc) setup_mutex, NULL);

	loop = g_main_loop_new (NULL, FALSE);
	g_mutex_lock (tag_read);
	setup_pipeline ();

	g_object_set (G_OBJECT (src), "location", path, NULL);

	/* Connect callback to identify audio and/or video tracks. */
	g_signal_connect (decode,
			  "new-decoded-pad",
			  G_CALLBACK (newpad_cb),
			  &has_video);

	g_signal_connect (decode,
			  "no-more-pads",
			  G_CALLBACK (no_more_pads_cb),
			  loop);

	/* Run main loop to allow decodebin to create pads. Quit on
	 * "no-more-pads" signal. */
	g_idle_add ((GSourceFunc) pause_pipeline, pipeline);
	g_timeout_add_seconds (1, (GSourceFunc) quit_mainloop, loop);
	g_main_loop_run (loop);

	fmt = GST_FORMAT_TIME;

	if (! gst_element_query_duration (typefind,
					 &fmt,
					 &nanoduration)) {
		g_warning ("Could not determine duration");
	} else {
		record->priv->duration = nanoduration / GST_SECOND;
	}

	if (!message_loop (GST_ELEMENT (pipeline), &tags)) {
		g_warning ("Failed in message reading for %s\n", path);
	}

	/* NOTE: Must set has_video before calling insert_tag. */
	g_warning ("%s %s", path, has_video ? "has video" : "doesn't have video");
	record->priv->has_video = has_video;

	if (tags) {
		gst_tag_list_foreach (tags, (GstTagForeachFunc) insert_tag, record);
		gst_tag_list_free (tags);
		tags = NULL;
	} else
		g_warning ("No metadata found for %s\n", path);

_return:
	gst_element_set_state (pipeline, GST_STATE_NULL);
	teardown_pipeline ();

	g_mutex_unlock (tag_read);
}

const char *dmapd_daap_record_get_location    (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->location;
}

const char *dmapd_daap_record_get_title       (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->title;
}

const char *dmapd_daap_record_get_album       (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->album;
}

const char *dmapd_daap_record_get_artist      (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->artist;
}

const char *dmapd_daap_record_get_genre       (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->genre;
}

const char *dmapd_daap_record_get_format      (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->format;
}

gint        dmapd_daap_record_get_filesize (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->filesize;
}

gint        dmapd_daap_record_get_rating (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->rating;
}

gint32      dmapd_daap_record_get_duration (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->duration;
}

gint32      dmapd_daap_record_get_track (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->track;
}

gint32      dmapd_daap_record_get_year (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->year;
}

gint32      dmapd_daap_record_get_firstseen (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->firstseen;
}

gint32      dmapd_daap_record_get_mtime (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->mtime;
}

gint32      dmapd_daap_record_get_disc (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->disc;
}

gulong      dmapd_daap_record_get_bitrate (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->bitrate;
}

gboolean    dmapd_daap_record_has_video (DAAPRecord *record)
{
	return DMAPD_DAAP_RECORD (record)->priv->has_video;
}

gboolean dmapd_daap_record_itunes_compat (DAAPRecord *record)
{
	const gchar *format = DMAPD_DAAP_RECORD (record)->priv->real_format;

	g_warning ("FORMAT: %s", format);

	if (! strcmp (format, "mp3"))
		return TRUE;
	else
		return FALSE;
}

void dmapd_daap_record_set_transcode_format (DAAPRecord *record, const gint fmt)
{
	DMAPD_DAAP_RECORD (record)->priv->transcode_format = fmt;
	DMAPD_DAAP_RECORD (record)->priv->format = g_strdup (g_gst_formats[DMAPD_DAAP_RECORD (record)->priv->transcode_format].extension);

	/* FIXME: free previous factory and turn into a seperate fn to be called after initial setting of transcode_format in ..._new? */
	switch (DMAPD_DAAP_RECORD (record)->priv->transcode_format) {
		case WAV16:
			DMAPD_DAAP_RECORD (record)->priv->factory = G_GST_INPUT_STREAM_FACTORY (g_gst_wav_input_stream_factory_new ());
			break;
		case MP3:
			DMAPD_DAAP_RECORD (record)->priv->factory = G_GST_INPUT_STREAM_FACTORY (g_gst_mp3_input_stream_factory_new ());
			break;
		case RAW:
		default:
			DMAPD_DAAP_RECORD (record)->priv->factory = G_GST_INPUT_STREAM_FACTORY (g_gst_raw_input_stream_factory_new ());
			break;
	}
}

GInputStream *dmapd_daap_record_read (DAAPRecord *record, GError **error)
{
	return G_INPUT_STREAM (g_gst_input_stream_factory_create (DMAPD_DAAP_RECORD (record)->priv->factory,  DMAPD_DAAP_RECORD (record)->priv->location));
}

static void dmapd_daap_record_init (DmapdDAAPRecord *record)
{
	record->priv = DMAPD_DAAP_RECORD_GET_PRIVATE (record);
}

static void dmapd_daap_record_class_init (DmapdDAAPRecordClass *klass)
{
	g_type_class_add_private (klass, sizeof (DmapdDAAPRecordPrivate));
}

static void dmapd_daap_record_interface_init (gpointer iface, gpointer data)
{
	DAAPRecordInterface *daap_record = iface;

	g_assert (G_TYPE_FROM_INTERFACE (daap_record) == TYPE_DAAP_RECORD);

	daap_record->get_location = dmapd_daap_record_get_location;
	daap_record->get_title = dmapd_daap_record_get_title;
	daap_record->get_filesize = dmapd_daap_record_get_filesize;
	daap_record->get_rating = dmapd_daap_record_get_rating;
	daap_record->get_format = dmapd_daap_record_get_format;
	daap_record->get_album = dmapd_daap_record_get_album;
	daap_record->get_artist = dmapd_daap_record_get_artist;
	daap_record->get_genre = dmapd_daap_record_get_genre;
	daap_record->get_duration = dmapd_daap_record_get_duration;
	daap_record->get_track = dmapd_daap_record_get_track;
	daap_record->get_year = dmapd_daap_record_get_year;
	daap_record->get_firstseen = dmapd_daap_record_get_firstseen;
	daap_record->get_mtime = dmapd_daap_record_get_mtime;
	daap_record->get_disc = dmapd_daap_record_get_disc;
	daap_record->get_bitrate = dmapd_daap_record_get_bitrate;
	daap_record->has_video = dmapd_daap_record_has_video;
	daap_record->itunes_compat = dmapd_daap_record_itunes_compat;
	daap_record->read = dmapd_daap_record_read;
}

G_DEFINE_TYPE_WITH_CODE (DmapdDAAPRecord, dmapd_daap_record, G_TYPE_OBJECT, 
			 G_IMPLEMENT_INTERFACE (TYPE_DAAP_RECORD, dmapd_daap_record_interface_init))

DmapdDAAPRecord *dmapd_daap_record_new (const char *path)
{
	const gchar *unknown = g_strdup ("Unknown");
	DmapdDAAPRecord *record = NULL;
	struct stat buf;
	char *format;

	record = DMAPD_DAAP_RECORD (g_object_new (TYPE_DMAPD_DAAP_RECORD, NULL));

	if (stat (path, &buf) == -1) {
		g_warning ("Unable to determine size of %s", path);
	} else {
		record->priv->filesize = buf.st_size;
	}

	record->priv->location = g_strdup_printf ("file://%s", path);
	record->priv->transcode_format = RAW;
	record->priv->factory = G_GST_INPUT_STREAM_FACTORY (g_gst_raw_input_stream_factory_new ());

	record->priv->title  = g_basename (path);  /* FIXME: possible leak */
	record->priv->artist = unknown;
	record->priv->album  = unknown;
	record->priv->genre  = unknown;
	record->priv->format = unknown;
	record->priv->real_format = determine_format (record, unknown);
	record->priv->track = 1;
	record->priv->duration = 1;

	read_metadata (record, path);

	record->priv->rating = 5;	/* FIXME */
	record->priv->year = 1984;	/* FIXME */
	record->priv->firstseen = 1;	/* FIXME */
	record->priv->mtime = 1;	/* FIXME */
	record->priv->disc = 1;		/* FIXME */
	record->priv->bitrate = 128;	/* FIXME */

	return record;
}
