/*
 * GGstInputStream class: Open a URI using g_gst_input_stream_new ().
 * Data is decoded using GStreamer and is then made available by the class's
 * read operations.
 *
 * Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <gst/app/gstappsink.h>
#include <gst/app/gstappbuffer.h>

#include "g-gst-input-stream.h"
#include "util.h"

#define GST_APP_MAX_BUFFERS 1024
#define DECODED_BUFFER_SIZE 1024 * 128
#define QUEUE_PUSH_WAIT_SECONDS 10
#define QUEUE_POP_WAIT_SECONDS 1

struct GGstInputStreamPrivate {
	GQueue *buffer;
	gsize read_request;		/* Size of data asked for */
	gsize write_request;		/* Number of bytes that must be read
					 * to make room for write */
	GCond *buffer_read_ready;	/* Signals when buffer >= read_req. */
	GCond *buffer_write_ready;	/* Signals when buffer not full. */
	GMutex *buffer_mutex; 		/* Protects buffer and read_request */
	gboolean buffer_closed;		/* May close before decoding complete */
	GstElement *pipeline;
	GstElement *src;
	GstElement *decode;
	GstElement *convert;
	GstCaps *filter;
	GstElement *encode;
	GstElement *sink;
};

static goffset
g_gst_input_stream_tell (GSeekable *seekable)
{
	/* FIXME: implement return current position in stream. */
	g_assert (FALSE);
}

static gboolean
g_gst_input_stream_can_seek (GSeekable *seekable)
{
	return TRUE;
}

static gboolean
g_gst_input_stream_seek (GSeekable *seekable,
			 goffset offset,
			 GSeekType type,
			 GCancellable *cacellable,
			 GError **error)
{
	GGstInputStream *stream;
	goffset absolute;

	stream = G_GST_INPUT_STREAM (seekable);

	switch (type) {
		/* FIXME: implement:
		 * case G_SEEK_CUR:
		 *	absolute = stream->priv->pos + offset;
		 *	break;
		 *
		 * case G_SEEK_END:
		 *	absolute = stream->priv->len + offset;
		 *	break;
		 */

		 case G_SEEK_SET:
		 	absolute = offset;
		 	break;

		default:
			g_set_error (error,
				     G_IO_ERROR,
				     G_IO_ERROR_INVALID_ARGUMENT,
				     "Invalid GSeekType supplied");

		return FALSE;
	}

	/* FIXME: implement:
	 * if (absolute < 0 || absolute > stream->priv->len) {
	 *	g_set_error_literal (error,
	 *		G_IO_ERROR,
	 *		G_IO_ERROR_INVALID_ARGUMENT,
	 *		_("Invalid seek request"));
	 *	return FALSE;
	 * }
	 */

	if (! gst_element_seek_simple (G_GST_INPUT_STREAM (seekable)->priv->pipeline,
				GST_FORMAT_BYTES,
				GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT,
				absolute)) {
		g_set_error (error,
			     G_IO_ERROR,
			     G_IO_ERROR_FAILED,
			     "Seek failed");
		return FALSE;
	}

	return TRUE;
}

static gboolean
g_gst_input_stream_can_truncate (GSeekable *seekable)
{
	return FALSE;
}

static gboolean
g_gst_input_stream_truncate (GSeekable *seekable,
			     goffset offset,
			     GCancellable *cancellable,
			     GError **error)
{
	g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
		     "Cannot truncate GGstInputStream");
	return FALSE;
}

static void
g_gst_input_stream_seekable_iface_init (GSeekableIface *iface)
{
	iface->tell         = g_gst_input_stream_tell;
	iface->can_seek     = g_gst_input_stream_can_seek;
	iface->seek         = g_gst_input_stream_seek;
	iface->can_truncate = g_gst_input_stream_can_truncate;
	iface->truncate_fn  = g_gst_input_stream_truncate;
}

static void
new_decoded_pad_cb (GstElement *element,
		    GstPad *pad,
		    gboolean arg,
		    GGstInputStream *stream)
{
	/* Link remaining pad after decodebin does its magic. */
	GstPad *conv_pad;

	conv_pad = gst_element_get_static_pad (stream->priv->convert, "sink");
	g_assert (conv_pad != NULL);

	if (pads_compatible (pad, conv_pad)) {
		g_assert (! GST_PAD_IS_LINKED (
			gst_element_get_static_pad (stream->priv->convert, "sink")));

		gst_pad_link (pad, conv_pad);

		if (stream->priv->filter) {
			if (gst_element_link_filtered (stream->priv->convert, 
						   stream->priv->encode, 
						   stream->priv->filter) == FALSE) {
				g_warning ("Error linking convert and encode elements");
			}
			if (gst_element_link (stream->priv->encode,
					  stream->priv->sink) == FALSE) {
				g_warning ("Error linking encode and sink elements");
			}
		} else {
			if (gst_element_link_many (stream->priv->convert,
						   stream->priv->encode,
						   stream->priv->sink,
						   NULL) == FALSE) {
				g_warning ("Error linking convert and sink elements");
			}
		}
	} else {
		g_warning ("Could not link GStreamer pipeline.");
	}
}

void
g_gst_input_stream_new_buffer_cb (GstElement *element, GGstInputStream *stream)
{
	gsize i;
	guint8 *ptr;
	GTimeVal time;
	GstBuffer *buffer;

	/* FIXME: Is this necessary? I am trying to protect against this
	 * thread manipulating data after the pipeline has been destroyed.
	 * see also g_gst_input_stream_close ().
	 */
	g_mutex_lock (stream->priv->buffer_mutex);

	if (stream->priv->buffer_closed) {
		g_warning ("Unread data");
		goto _return;
	}

	g_get_current_time (&time);
	g_time_val_add (&time, QUEUE_PUSH_WAIT_SECONDS * 1000000);

	buffer = gst_app_sink_pull_buffer (GST_APP_SINK (element));

	if (buffer == NULL) {
		g_warning ("Error getting GStreamer buffer");
		goto _return;
	}

	/* FIXME: this actually allows buffer to grow larger than max. */
	if (g_queue_get_length (stream->priv->buffer) + GST_BUFFER_SIZE (buffer)> DECODED_BUFFER_SIZE) {
		stream->priv->write_request = GST_BUFFER_SIZE (buffer);
		if (! g_cond_timed_wait (stream->priv->buffer_write_ready,
			   stream->priv->buffer_mutex, &time)) {
			g_warning ("Timeout waiting for buffer to empty; will drop");
		}
		/* Required again because g_cond_timed_wait released mutex. */
		if (stream->priv->buffer_closed) {
			g_warning ("Unread data");
			goto _return;
		}
	} else {
		stream->priv->write_request = 0;
	}

	if (g_queue_get_length (stream->priv->buffer) + GST_BUFFER_SIZE (buffer) <= DECODED_BUFFER_SIZE) {
		ptr = GST_BUFFER_DATA (buffer);

		for (i = 0; i < GST_BUFFER_SIZE (buffer); i++) {
			g_queue_push_tail (stream->priv->buffer, GINT_TO_POINTER((gint) *ptr++));
		}
	}

	gst_buffer_unref (buffer);

	if (g_queue_get_length (stream->priv->buffer)
	    >= stream->priv->read_request) {
		stream->priv->read_request = 0;
		g_cond_signal (stream->priv->buffer_read_ready);
	}

_return:
	g_mutex_unlock (stream->priv->buffer_mutex);
}

GInputStream* g_gst_input_stream_new (const gchar *uri,
				      const gint transcode_format)
{
	GstStateChangeReturn sret;
	GstState state;
	GGstInputStream *stream;
	GError *error;

	stream = G_GST_INPUT_STREAM (g_object_new (TYPE_G_GST_INPUT_STREAM,
						   NULL));

#if 0
	stream->priv->pipeline = gst_pipeline_new ("pipeline");

	stream->priv->src     = gst_element_factory_make ("giosrc", "src");
	switch (transcode_format) {
		case RAW:
			g_warning ("RAW"); /* No transcoding. */
			stream->priv->decode  = gst_element_factory_make ("queue", "decode");
			stream->priv->convert = gst_element_factory_make ("queue", "convert");
			stream->priv->filter = NULL;
			stream->priv->encode  = gst_element_factory_make ("queue", "encode");
			break;
		case WAV8:
			stream->priv->decode  = gst_element_factory_make ("decodebin", "decode");
			stream->priv->convert = gst_element_factory_make ("audioconvert", "convert");
			/* Roku clients support a subset of the WAV format. */
			stream->priv->filter = gst_caps_new_simple ("audio/x-raw-int",
								    "channels", G_TYPE_INT, 2,
								    "width", G_TYPE_INT, 8,
								    "depth", G_TYPE_INT, 8,
								    NULL);
			stream->priv->encode  = gst_element_factory_make ("wavenc", "encode");
			break;
		case WAV16:
			stream->priv->decode  = gst_element_factory_make ("decodebin", "decode");
			stream->priv->convert = gst_element_factory_make ("audioconvert", "convert");
			/* Roku clients support a subset of the WAV format. */
			stream->priv->filter = gst_caps_new_simple ("audio/x-raw-int",
								    "channels", G_TYPE_INT, 2,
								    "width", G_TYPE_INT, 16,
								    "depth", G_TYPE_INT, 16,
								    NULL);
			stream->priv->encode  = gst_element_factory_make ("wavenc", "encode");
			break;
		case MP3:
			g_warning ("MP3");
			stream->priv->decode  = gst_element_factory_make ("decodebin", "decode");
			stream->priv->convert = gst_element_factory_make ("audioconvert", "convert");
			stream->priv->filter = NULL;
			stream->priv->encode  = gst_element_factory_make ("lame", "encode");
			break;
		default:
			g_error ("Unknown transcode format");
	}

	stream->priv->sink = gst_element_factory_make ("appsink", "sink");

	gst_bin_add_many (GST_BIN (stream->priv->pipeline),
			  stream->priv->src,
			  stream->priv->decode,
			  stream->priv->convert,
			  stream->priv->encode,
			  stream->priv->sink,
			  NULL);

	if (gst_element_link_many (stream->priv->src,
			       stream->priv->decode,
			       NULL) == FALSE) {
		g_warning ("Error linking source and decode elements");
	}

	g_object_set (G_OBJECT (stream->priv->src), "location", uri, NULL);

	switch (transcode_format) {
		case RAW:
			if (gst_element_link_many (stream->priv->decode,
						   stream->priv->convert,
						   stream->priv->encode,
						   stream->priv->sink,
						   NULL) == FALSE) {
				g_warning ("Error linking source and decode elements");
	}
			break;
		case WAV8:
		case WAV16:
			g_signal_connect (stream->priv->decode, "new-decoded-pad", G_CALLBACK (new_decoded_pad_cb), stream);
			break;
		case MP3:
			/* quality=9 is important for fast, realtime transcoding: */
			g_object_set (G_OBJECT (stream->priv->encode), "quality", 9, NULL);
			g_object_set (G_OBJECT (stream->priv->encode), "bitrate", 128, NULL);
			g_object_set (G_OBJECT (stream->priv->encode), "vbr", 0, NULL);
			g_signal_connect (stream->priv->decode, "new-decoded-pad", G_CALLBACK (new_decoded_pad_cb), stream);
			break;
	}

	g_object_set (G_OBJECT (stream->priv->sink), "emit-signals", TRUE, "sync", FALSE, NULL);
	gst_app_sink_set_max_buffers (GST_APP_SINK (stream->priv->sink),
				      GST_APP_MAX_BUFFERS);
	gst_app_sink_set_drop (GST_APP_SINK (stream->priv->sink), FALSE);

	g_signal_connect (stream->priv->sink, "new-buffer", G_CALLBACK (g_gst_input_stream_new_buffer_cb), stream);

	/* FIXME: this technique is shared with dmapd-daap-share.c */
	sret = gst_element_set_state (stream->priv->pipeline, GST_STATE_PLAYING);
	if (GST_STATE_CHANGE_ASYNC == sret) {
		if (GST_STATE_CHANGE_SUCCESS != gst_element_get_state (GST_ELEMENT (stream->priv->pipeline), &state, NULL, 5 * GST_SECOND)) {
			g_warning ("State change failed for %s.", uri);
		}
	} else if (sret != GST_STATE_CHANGE_SUCCESS) {
		g_print ("%s - Could not read file %d\n", uri, sret);
	}
#endif

	g_assert (G_IS_SEEKABLE (stream));
	return G_INPUT_STREAM (stream);
}

static gssize
min (gssize a, gssize b)
{
	return a < b ? a : b;
}

static gssize
g_gst_input_stream_read (GInputStream  *stream,
			 void          *buffer,
			 gsize          count,
			 GCancellable  *cancellable,
			 GError       **error)
{
	int i;
	GGstInputStream *gst_stream = G_GST_INPUT_STREAM (stream);
	GTimeVal time;

	g_get_current_time (&time);
	g_time_val_add (&time, QUEUE_POP_WAIT_SECONDS * 1000000);

	g_mutex_lock (gst_stream->priv->buffer_mutex);

	gst_stream->priv->read_request = count;
	if (g_queue_get_length (gst_stream->priv->buffer) < count
	    && ! g_cond_timed_wait (gst_stream->priv->buffer_read_ready,
			   gst_stream->priv->buffer_mutex,
			   &time)) {
		/* Timeout: Count is now what's remaining.  Let's hope
		 * we have enough of a lead on encoding so that this one
		 * second timeout will go unnoticed.
		 */
		g_warning ("Timeout waiting for converted data");
		/* Depending on timing, more data may have been written
		 * since check: do not pull more than count:
		 */
		count = min (count, g_queue_get_length (
					gst_stream->priv->buffer));
	}

	for (i = 0; i < count; i++) {
		((guint8 *) buffer)[i] = GPOINTER_TO_INT (
			g_queue_pop_head (gst_stream->priv->buffer));
	}

	if (gst_stream->priv->write_request > count)
		gst_stream->priv->write_request -= count;
	else
		gst_stream->priv->write_request = 0;

	if (gst_stream->priv->write_request <= 0) {
		g_cond_signal (gst_stream->priv->buffer_write_ready);
	}

	g_mutex_unlock (gst_stream->priv->buffer_mutex);

	return count;
}

static gssize
g_gst_input_stream_skip (GInputStream *stream,
			    gsize count,
			    GCancellable *cancellable,
			    GError **error)
{
}

static gboolean
g_gst_input_stream_close (GInputStream *stream,
			  GCancellable *cancellable,
			  GError **error)
{
	GGstInputStream *gst_stream = G_GST_INPUT_STREAM (stream);

	gst_element_set_state (gst_stream->priv->pipeline, GST_STATE_NULL);
	gst_object_unref (GST_OBJECT (gst_stream->priv->pipeline));
	if (gst_stream->priv->filter)
		gst_caps_unref (GST_CAPS (gst_stream->priv->filter));

	g_mutex_lock (gst_stream->priv->buffer_mutex);

	g_queue_free (gst_stream->priv->buffer);
	gst_stream->priv->buffer_closed = TRUE;

	g_mutex_unlock (gst_stream->priv->buffer_mutex);
}

static gssize
g_gst_input_stream_read_finish (GInputStream *stream,
				GAsyncResult *result,
				GError **error)
{
}

static gssize
g_gst_input_stream_skip_finish (GInputStream *stream,
				GAsyncResult *result,
				GError **error)
{
}

static void
g_gst_input_stream_close_async (GInputStream *stream,
				int io_priority,
				GCancellable *cancellabl,
				GAsyncReadyCallback callback,
				gpointer data)
{
}

static void
g_gst_input_stream_read_async (GInputStream *stream,
			       void *buffer,
			       gsize count,
			       int io_priority,
			       GCancellable *cancellable,
			       GAsyncReadyCallback callback,
			       gpointer user_data)
{
}

static void
g_gst_input_stream_skip_async (GInputStream *stream,
			       gsize count,
			       int io_priority,
			       GCancellable *cancellabl,
			       GAsyncReadyCallback callback,
			       gpointer datae)
{
}

static gboolean
g_gst_input_stream_close_finish (GInputStream *stream,
			         GAsyncResult *result,
				 GError **error)
{
}

static void
g_gst_input_stream_class_init (GGstInputStreamClass *klass)
{
	GObjectClass *object_class;
	GInputStreamClass *istream_class;

	g_type_class_add_private (klass, sizeof (GGstInputStreamPrivate));

	istream_class = G_INPUT_STREAM_CLASS (klass);
	istream_class->read_fn      = g_gst_input_stream_read;
	istream_class->skip         = g_gst_input_stream_skip;
	istream_class->close_fn     = g_gst_input_stream_close;
	istream_class->read_async   = g_gst_input_stream_read_async;
	istream_class->read_finish  = g_gst_input_stream_read_finish;
	istream_class->skip_async   = g_gst_input_stream_skip_async;
	istream_class->skip_finish  = g_gst_input_stream_skip_finish;
	istream_class->close_async  = g_gst_input_stream_close_async;
	istream_class->close_finish = g_gst_input_stream_close_finish;
}

static void
g_gst_input_stream_init (GGstInputStream *stream)
{
	stream->priv = G_GST_INPUT_STREAM_GET_PRIVATE (stream);

	stream->priv->buffer = g_queue_new ();
	stream->priv->read_request = 0;
	stream->priv->write_request = 0;
	stream->priv->buffer_read_ready = g_cond_new ();
	stream->priv->buffer_write_ready = g_cond_new ();
	stream->priv->buffer_mutex = g_mutex_new ();
	stream->priv->buffer_closed = FALSE;
}

G_DEFINE_TYPE_WITH_CODE (GGstInputStream, g_gst_input_stream,
			 G_TYPE_INPUT_STREAM,
			 G_IMPLEMENT_INTERFACE (G_TYPE_SEEKABLE,
				g_gst_input_stream_seekable_iface_init))
