/*
 * GGstRAWInputStream class: Open a URI using g_gst_raw_input_stream_new ().
 * Data is provided in its original format by the class's read operations.
 *
 * Copyright (C) 2009 W. Michael Petullo <mike@flyn.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <gst/app/gstappsink.h>

#include "g-gst-raw-input-stream.h"

#define GST_APP_MAX_BUFFERS 1024

struct GGstRAWInputStreamPrivate {
	GstElement *pipeline;
	GstElement *src;
	GstElement *sink;
};

GInputStream* g_gst_raw_input_stream_new (const gchar *uri)
{
	GstStateChangeReturn sret;
	GstState state;
	GGstRAWInputStream *stream;
	GError *error;

	stream = G_GST_RAW_INPUT_STREAM (g_object_new (TYPE_G_GST_RAW_INPUT_STREAM,
						   NULL));

	stream->priv->pipeline = gst_pipeline_new ("pipeline");

	stream->priv->src     = gst_element_factory_make ("giosrc", "src");
	stream->priv->sink    = gst_element_factory_make ("appsink", "sink");

	gst_bin_add_many (GST_BIN (stream->priv->pipeline),
			  stream->priv->src,
			  stream->priv->sink,
			  NULL);

	if (gst_element_link (stream->priv->src, stream->priv->sink) == FALSE) {
		g_warning ("Error linking source and sink elements");
	}

	g_object_set (G_OBJECT (stream->priv->src), "location", uri, NULL);

	g_object_set (G_OBJECT (stream->priv->sink), "emit-signals", TRUE, "sync", FALSE, NULL);
	gst_app_sink_set_max_buffers (GST_APP_SINK (stream->priv->sink),
				      GST_APP_MAX_BUFFERS);
	gst_app_sink_set_drop (GST_APP_SINK (stream->priv->sink), FALSE);

	g_signal_connect (stream->priv->sink, "new-buffer", G_CALLBACK (g_gst_input_stream_new_buffer_cb), stream);

	/* FIXME: this technique is shared with dmapd-daap-share.c */
	sret = gst_element_set_state (stream->priv->pipeline, GST_STATE_PLAYING);
	if (GST_STATE_CHANGE_ASYNC == sret) {
		if (GST_STATE_CHANGE_SUCCESS != gst_element_get_state (GST_ELEMENT (stream->priv->pipeline), &state, NULL, 5 * GST_SECOND)) {
			g_warning ("State change failed for %s.", uri);
		}
	} else if (sret != GST_STATE_CHANGE_SUCCESS) {
		g_print ("%s - Could not read file %d\n", uri, sret);
	}

	g_assert (G_IS_SEEKABLE (stream));
	return G_INPUT_STREAM (stream);
}

G_DEFINE_TYPE (GGstRAWInputStream, g_gst_raw_input_stream, TYPE_G_GST_INPUT_STREAM)

static void
g_gst_raw_input_stream_class_init (GGstRAWInputStreamClass *klass)
{
	g_type_class_add_private (klass, sizeof (GGstRAWInputStreamPrivate));
}

static void
g_gst_raw_input_stream_init (GGstRAWInputStream *stream)
{
	stream->priv = G_GST_RAW_INPUT_STREAM_GET_PRIVATE (stream);

}
